import argparse
import contextlib
import os
import re
import sys

from doq import __version__
from doq.outputter import (
    JSONOutputter,
    StringOutptter,
)
from doq.parser import parse
from doq.template import Template


def _sort(key):
    return key['start_lineno']


def find_files(basedir):
    r = re.compile(r'.*.py$')
    files = []
    for root, directories, children in os.walk(os.path.abspath(basedir)):
        directories[:] = [d for d in directories if not d[0] == '.']
        if len(children):
            ret = list(filter(r.match, children))
            if len(ret):
                files += [os.path.join(root, r) for r in ret]

    return files


def get_lines(file, start, end):
    lines = []
    with contextlib.closing(file) as f:
        lines = [line.strip('\n') for line in f]
        start = start - 1
        end = len(lines) if end == 0 else end

        lines = lines[start:end]

    return lines


def get_template_path(template_path, formatter):
    if not template_path:
        path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'templates',
        )

        return os.path.join(os.path.abspath(path), formatter)

    return os.path.abspath(template_path)


def generate_def_docstrings(signature, template, is_exception=False, is_yield=False):
    docstrings = []
    for d in signature['defs']:
        if d['is_doc_exists'] is False:
            filename = 'noarg.txt'
            if ('params' in d and len(d['params'])) \
                    or ('return_type' in d and d['return_type']) \
                    or (is_exception and 'exceptions' in d and len(d['exceptions']) > 0) \
                    or (is_yield and 'yields' in d and len(d['yields']) > 0):
                filename = 'def.txt'
            elif 'defs' in d:
                filename = 'class.txt'

            docstring = template.load(params=d, filename=filename)
            docstrings.append(
                {
                    'docstring': docstring,
                    'start_lineno': d['start_lineno'],
                    'start_col': d['start_col'],
                    'end_lineno': d['end_lineno'],
                    'end_col': d['start_col'],
                },
            )
            if 'defs' in d:
                results = generate_def_docstrings(d, template, is_exception, is_yield)
                if len(results):
                    docstrings += results

    return docstrings


def is_exception_enabled(path):
    with open(path) as f:
        return 'exceptions' in f.read()


def is_yield_enabled(path):
    with open(path) as f:
        return 'yields' in f.read()


def generate_docstrings(
    code,
    path,
    omissions=None,
    ignore_exception=False,
    ignore_yield=False,
):
    template = Template(paths=[path])
    signatures = parse(
        '\n'.join(code),
        omissions=omissions,
        ignore_exception=ignore_exception,
        ignore_yield=ignore_yield,
    )
    is_exception = False if ignore_exception else is_exception_enabled(os.path.join(path, 'def.txt'))
    is_yield = False if ignore_yield else is_yield_enabled(os.path.join(path, 'def.txt'))

    docstrings = []
    for signature in signatures:
        if 'defs' in signature:
            # Class docstring
            if signature['is_doc_exists'] is False:
                docstring = template.load(params=signature, filename='class.txt')
                docstrings.append(
                    {
                        'docstring': docstring,
                        'start_lineno': signature['start_lineno'],
                        'start_col': signature['start_col'],
                        'end_lineno': signature['end_lineno'],
                        'end_col': signature['end_col'],
                    },
                )

            # Method docstring
            docstrings += generate_def_docstrings(signature, template, is_exception, is_yield)
        else:
            if signature['is_doc_exists'] is False:
                filename = 'noarg.txt'
                if len(signature['params']) \
                        or signature['return_type'] \
                        or is_exception and signature['exceptions'] \
                        or is_yield and signature['yields']:
                    filename = 'def.txt'

                # Function docstring
                docstring = template.load(params=signature, filename=filename)
                docstrings.append(
                    {
                        'docstring': docstring,
                        'start_lineno': signature['start_lineno'],
                        'start_col': signature['start_col'],
                        'end_lineno': signature['end_lineno'],
                        'end_col': signature['start_col'],
                    },
                )

    docstrings.sort(key=_sort)

    return docstrings


def get_targets(args):
    targets = []
    if args.recursive:
        files = find_files(args.directory)
        for file in files:
            with open(file) as f:
                lines = get_lines(f, args.start, args.end)
                if len(lines) == 0:
                    continue
                targets.append({
                    'path': file,
                    'lines': lines,
                })
    else:
        lines = get_lines(args.file, args.start, args.end)
        if len(lines) == 0:
            return False

        args.file.name == '<stdin>'
        path = args.file.name \
            if args.file.name == '<stdin>' \
            else os.path.abspath(args.file.name)

        targets.append({
            'path': path,
            'lines': lines,
        })

    return targets


def run(args):
    targets = get_targets(args)
    path = get_template_path(
        template_path=args.template_path,
        formatter=args.formatter,
    )

    if not os.path.exists(path):
        return False

    omissions = args.omit.split(',') if args.omit else None

    if not targets:
        return

    for target in targets:
        docstrings = generate_docstrings(
            code=target['lines'],
            path=path,
            omissions=omissions,
            ignore_exception=args.ignore_exception,
            ignore_yield=args.ignore_yield,
        )
        if len(docstrings) == 0:
            continue

        if args.style == 'json':
            outputter = JSONOutputter()
        else:
            outputter = StringOutptter()

        output = outputter.format(
            lines=target['lines'],
            docstrings=docstrings,
            indent=args.indent,
        )

        if args.write and target['path'] != '<stdin>':
            with open(target['path'], 'w') as f:
                f.write(output + '\n')

        else:
            sys.stdout.write(output + '\n')

    return True


def parse_options():
    parser = argparse.ArgumentParser(
        prog='doq',
        description='Docstring generator.',
        add_help=True,
    )
    parser.add_argument(
        '-f',
        '--file',
        type=argparse.FileType('r'),
        default='-',
        help='File or STDIN',
    )
    parser.add_argument(
        '--start',
        type=int,
        default=1,
        help='Start lineno',
    )
    parser.add_argument(
        '--end',
        type=int,
        default=0,
        help='End lineno',
    )
    parser.add_argument(
        '-t',
        '--template_path',
        type=str,
        default=None,
        help='Path to template directory',
    )
    parser.add_argument(
        '-s',
        '--style',
        type=str,
        default='string',
        help='Output style string or json',
    )
    parser.add_argument(
        '--formatter',
        type=str,
        default='sphinx',
        help='Docstring formatter. sphinx,google or numpy',
    )
    parser.add_argument(
        '--indent',
        type=int,
        default=4,
        help='Indent number',
    )
    parser.add_argument(
        '--omit',
        type=str,
        default=None,
        help='Omit first argument such as self',
    )
    parser.add_argument(
        '-r',
        '--recursive',
        action='store_true',
        help='Run recursively over directories',
    )
    parser.add_argument(
        '-d',
        '--directory',
        default='',
        help='Dire',
    )
    parser.add_argument(
        '-w',
        '--write',
        action='store_true',
        help='Edit files in-place',
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version='%(prog)s {0}'.format(__version__),
        help='Output the version number',
    )
    parser.add_argument(
        '--ignore_exception',
        action='store_true',
        help='Ignore exception statements',
    )
    parser.add_argument(
        '--ignore_yield',
        action='store_true',
        help='Ignore yield statements',
    )

    args = parser.parse_args()

    return args


def main():
    args = parse_options()
    try:
        ret = run(args)
        if ret:
            sys.exit(0)

        sys.exit(1)
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()



import re

import parso


def sort_by_lineno(key):
    if len(key['defs']):
        return key['defs'][0]['start_lineno']

    return key['start_lineno']


def get_return_type(line):
    matched = re.search(r'\)(.*)->(.*):', line)
    if matched:
        return_type = matched.group()[1:]
        return_type = re.sub(r':$', '', return_type).strip()
        return_type = return_type.replace('->', '').strip()
        return return_type

    return None


def get_start_col(line, start_col):
    lines = line.split('\n')
    for v in lines:
        if v == '':
            # Maybe line break only row
            continue

        index = v.find('def')
        if index == start_col:
            return start_col

    # Maybe async keyword exists.
    # parso does not parse by parso and it's known issue.
    return start_col - 6  # 'async '


def parse_return_type(code, start_lineno, end_lineno):
    lines = code.strip().split('\n')
    lineno = end_lineno - start_lineno
    if lineno == 0:
        # Signature end
        return get_return_type(lines[0])

    for line in lines[0:lineno + 1]:
        return_type = get_return_type(line)
        if return_type:
            return return_type

    return None


def parse_defs(module, omissions=None, ignore_exception=False, ignore_yield=False):  # noqa C901
    if omissions is None:
        omissions = []

    results = []
    for d in module.iter_funcdefs():
        is_doc_exists = True if d.get_doc_node() else False

        (start_lineno, start_col) = d.start_pos
        (end_lineno, end_col) = d.end_pos

        code = d.get_code()
        start_col = get_start_col(code, start_col)

        name = d.name.value
        params = []
        is_classmethod = any('@classmethod' in d.get_code() for d in d.get_decorators())

        for i, p in enumerate(d.get_params()):
            if is_classmethod and i == 0:
                # Ignore first argument if method is `@classmethod`.
                continue

            if p.name.value in omissions and i == 0:
                # Method's first variable is maybe `self`.
                continue

            arguments = {'argument': None, 'annotation': None, 'default': None}
            arguments['argument'] = p.name.value
            if p.annotation:
                arguments['annotation'] = p.annotation.get_code().strip()
            if p.default:
                arguments['default'] = p.default.get_code().strip()

            params.append(arguments)

        return_type = None
        if d.children[3].value == '->':
            return_type = d.children[4].get_code().strip()

        yields = []
        if ignore_yield is False:
            for y in d.iter_yield_exprs():
                yields.append(y.children[1].get_first_leaf().value)

        exceptions = []
        if ignore_exception is False:
            for e in d.iter_raise_stmts():
                exceptions.append(e.children[1].get_first_leaf().get_code().strip())

        results.append(
            {
                'name': name,
                'params': params,
                'return_type': return_type,
                'start_lineno': start_lineno,
                'start_col': start_col,
                'end_lineno': end_lineno,
                'end_col': end_col,
                'is_doc_exists': is_doc_exists,
                'exceptions': exceptions,
                'yields': yields,
            },
        )

        nested = parse_defs(d, ignore_exception=ignore_exception, ignore_yield=ignore_yield)
        if len(nested):
            results += nested

        nested = parse_classdefs(d)
        if len(nested):
            results += nested

    return results


def parse_classdefs(module, ignore_exception=False, ignore_yield=False):
    results = []

    for c in module.iter_classdefs():
        is_doc_exists = True if c.get_doc_node() else False

        (start_lineno, start_col) = c.start_pos
        (end_lineno, end_col) = c.end_pos

        name = c.name.value
        defs = parse_defs(c, omissions=['self'], ignore_exception=ignore_exception, ignore_yield=ignore_yield)
        results.append(
            {
                'name': name,
                'defs': defs,
                'start_lineno': start_lineno,
                'start_col': start_col,
                'end_lineno': end_lineno,
                'end_col': end_col,
                'is_doc_exists': is_doc_exists,
            },
        )

        nested = parse_classdefs(c, ignore_exception=ignore_exception, ignore_yield=ignore_yield)
        if len(nested):
            results += nested

    results.sort(key=sort_by_lineno)

    return results


def parse(code, omissions=None, ignore_exception=False, ignore_yield=False):
    m = parso.parse(code)
    results = []
    if 'class' in code:
        results = parse_classdefs(m, ignore_exception=ignore_exception, ignore_yield=ignore_yield)

    results += parse_defs(
        m,
        omissions=omissions,
        ignore_exception=ignore_exception,
        ignore_yield=ignore_yield,
    )

    return results



from jinja2 import (
    Environment,
    FileSystemLoader,
)


class Template:
    def __init__(self, paths):
        self.env = Environment(
            loader=FileSystemLoader(paths),
            autoescape=False,
            auto_reload=False,
        )

    def load(self, params, filename=None):
        filename = filename or 'def.txt'
        template = self.env.get_template(filename)
        return template.render(**params)



import os

from setuptools import (
    find_packages,
    setup,
)

version = ''
with open('doq/__init__.py') as f:
    for line in f.readlines():
        if line.startswith('__version__'):
            version = line.split('=')[1].strip().replace("'", '')

rst_path = os.path.join(os.path.dirname(__file__), 'README.rst')
description = ''
with open(rst_path) as f:
    description = f.read()

setup(
    name='doq',
    version=version,
    author='Shinya Ohyanagi',
    author_email='sohyanagi@gmail.com',
    url='http://github.com/heavenshell/py-doq',
    description='Docstring generator',
    long_description=description,
    license='BSD',
    platforms='any',
    packages=find_packages(exclude=['tests']),
    package_dir={'': '.'},
    include_package_data=True,
    package_data={'doq': [
        'templates/google/*.txt',
        'templates/numpy/*.txt',
        'templates/sphinx/*.txt',
    ]},
    install_requires=['parso', 'jinja2'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Documentation',
        'Topic :: Software Development :: Documentation',
    ],
    entry_points="""
    [console_scripts]
    doq = doq.cli:main
    """,
    tests_require=['parameterized'],
    test_suite='tests',
)



import json
import re


class BaseOutputter:
    def format(self, lines, docstrings, indent=None):
        raise NotImplementedError()


class StringOutptter(BaseOutputter):
    def detect_insert_point(self, lines, start, end):
        if lines[start - 1].endswith(':') or re.search(r'\):', lines[start - 1]):
            # Found end of signature
            return start
        else:
            for i, line in enumerate(lines[start:end]):
                if line.endswith('):'):
                    # Found end of signature without type
                    return start + i + 1
                elif re.search(r'\):', line):
                    return start + i + 1
                elif re.search(r'\]:', line):
                    # Found end of signature type
                    #   def foo(a, b) -> Tuple[
                    #       int,
                    #   ]:
                    #       pass
                    return start + i + 1
                elif re.search(r'->(.*):', line):
                    # Found end of signature with type
                    return start + i + 1

        return start

    def format(self, lines, docstrings, indent=None):
        for d in reversed(docstrings):
            col = d['start_col'] + indent
            ret = []
            for line in d['docstring'].split('\n'):
                if line == '':
                    ret.append('')
                else:
                    ret.append('{0}{1}'.format(' ' * col, line))

            lineno = self.detect_insert_point(
                lines,
                d['start_lineno'],
                d['end_lineno'],
            )

            lines.insert(lineno, '\n'.join(ret))

        return '\n'.join(lines)


class JSONOutputter(BaseOutputter):
    def format(self, lines, docstrings, indent=None):
        results = []
        for d in docstrings:
            col = d['start_col'] + indent
            ret = []
            for line in d['docstring'].split('\n'):
                if line == '':
                    ret.append('')
                else:
                    ret.append('{0}{1}'.format(' ' * col, line))

            results.append({
                'docstring': '\n'.join(ret),
                'start_col': col,
                'start_lineno': d['start_lineno'],
                'end_col': d['end_col'],
                'end_lineno': d['end_lineno'],
            })

        return json.dumps(results)
